package HomeMade.Tracking;
import ij.IJ.*;
import java.util.Arrays;
import java.util.List;
import java.lang.Math;
import java.util.ArrayList;
import HomeMade.Tracking.*;

import ij.*;
public class FociTrackingMT {
	public static void trackmyFoci(data Data){
		double[] area = Data.area; // get all areas of all objects
		double[] x_coord = Data.x; // get all xpoints
		double[] y_coord = Data.y; // all ypoints
		double[] frame = Data.frame; // all frame numbers
		String ColorSet=Data.ColorSet; // the color of the drawn tracks
		ImagePlus imp =Data.imp; // the image we a re working with
		double x_coordmax = 0; // initiate the max x & y position at 0
		double y_coordmax = 0;
		
		//---Find the maximum x and y position
		for (int i=0; i<x_coord.length; i++){
			if (x_coord[i]>x_coordmax){
				x_coordmax=x_coord[i];
			}
			if (y_coord[i]>y_coordmax){
				y_coordmax=y_coord[i];
			}
		}
		
		//--Create the boundary sizes for double and tripple foci---
		double areasum = 0;
		for (int i=0; i<area.length; i++){
			areasum = areasum + area[i];
		}
		double two_foci = 1.5*areasum/area.length;
		double three_foci = 1.5*two_foci;

		double L=Math.ceil(Math.max(x_coordmax,y_coordmax)); // Take the maximum coordinate for determining pixel steps
		int delx = (int)L; // create space at maximum position size
		int dely = (int)L;
		space s = new space(delx,dely);
		
		int startiter = (int) frame[0];
		int enditer = (int) frame[frame.length-1];
		int rangesize = Data.trackRange;
		int framerange=4; // for looking over multiple stacks
		int objnum = 0;
		ObjectArray OA = new ObjectArray(enditer);
		
		OAindices OAi = new OAindices(enditer);
		Integer[] indicestrial = new Integer[0];
		// Create the first space
		
		while(frame[objnum] == startiter){ // load all objects of the first frame
			int ii = (int)Math.floor(x_coord[objnum]);
			int jj = (int)Math.floor(y_coord[objnum]);
			FociObject Foci = new FociObject(objnum+1, x_coord[objnum], y_coord[objnum], frame[objnum]);
			s.addtospace(ii, jj, Foci);
			OA.add(startiter-1, Foci);
			OAi.add(startiter-1, Foci.FOindex);
			objnum++;
			indicestrial = OAi.find(0, Foci.FOindex);
			
		}
		// Resize OA
		OA.resize(startiter-1);
		OAi.resize(startiter-1);
		// Add these objects to the correct frame (startiter-1)

		for (int iter=startiter+1; iter<=enditer; iter++){ // run through all subsequent frames

			List<FociObject> toaddtospace = new ArrayList<FociObject>(0); // make a new list of objects to add to the foci space
			int frameN=0;
			while (frame[objnum]==iter-frameN){ // check only those objects that are in the iter frame
				FociObject CurrentFoci = new FociObject(objnum+1, x_coord[objnum], y_coord[objnum], frame[objnum]); // make a new foci object
				int ii = (int)Math.floor(x_coord[objnum]); //get coordinates and round down
				int jj = (int)Math.floor(y_coord[objnum]); //get coordinates and round down
				//IJ.log("iter: "+iter+", x: "+ii+", y: "+jj);
				CurrentFoci.setcoords(ii,jj); // set integer coordinates
				toaddtospace.add(CurrentFoci); //add it to the space
				int neighbour = 0; // no neighbour yet
				int firsttry = 1; //start with looking 1 around
				int spaceindex = 1;
				int range = 1;
				DistanceVector[] dist;
				dist = new DistanceVector[10];
				//double[] distvalues;
				//distvalues = new double[10];
				int distsize = 0;
				ArrayList<Neighbour> distancelist = new ArrayList<Neighbour>(); // make a neighbour list
				
				while (neighbour==0){ // as long as this foci does not have a neigbour
					//IJ.log("F"+firsttry+" R:"+rangesize);
					if(firsttry<=rangesize){
						range = firsttry;
						spaceindex = 1;
					}else{
						IJ.log(""+frameN);
						// not sure if this does anything?
						/*if((firsttry>rangesize)&&(spaceindex<(iter-1))&&(spaceindex<3)){
							range = (firsttry % rangesize);
							spaceindex = (int)Math.floor((firsttry-1)/rangesize);
							IJ.log("SPI"+spaceindex);
						}else{
							//neighbour = 1;*/
							if (frameN<framerange){
								frameN+=1;
							} else {
								break;
							}
						//}
					}
					for (int aa=(ii-range);aa<=(ii+range);aa++){ // go from x-range to x + range
						for (int bb=(jj-range);bb<=(jj+range);bb++){ // go from y-range to y+range
							if (aa >= 0 && bb >= 0 && aa < delx && bb < dely){ // if aa and bb fall within the maximum image size
								ArrayList<FociObject> xlist = s.get(aa, bb); // get the current space list on this position
								int xx = xlist.size(); // if there are foci in this position
								for (int qq=0;qq<xx;qq++){ // run through the list of spaces
									FociObject F = xlist.get(qq);
									IJ.log("f"+F.getfr()+",cf "+CurrentFoci.getfr());
									if (F.getfr()==CurrentFoci.getfr()-frameN){
										//IJ.log("x1:"+x_coord[objnum]+", x2:"+F.x_coord);
										//IJ.log("y1:"+y_coord[objnum]+", y2:"+F.y_coord);
										double del_x = (x_coord[objnum]-F.x_coord); // determine dX
										double del_y = (y_coord[objnum]-F.y_coord); // determine dY
										double distance = Math.sqrt((del_x*del_x)+(del_y*del_y));// determine distance
										Neighbour N = new Neighbour(F,distance); //add it to the neighbour list
										distancelist.add(N); // if there are more neighbours at the same x, y borders they will all be added
										neighbour=1; // mark that there is a neighbour
									}
								}
							}
						}
					}
					firsttry++; // increase the range to look into
				}
				//objnum++;
				if (neighbour==1){ // There is a match!
					Arrays.sort(distancelist.toArray()); // sort the potential objects on distance
					distsize = distancelist.size();
					FociObject[] posobj = new FociObject[0];

					if (distsize>1){ // if there are more objects
						if (area[objnum]<two_foci){ // check if they could be 2 objects
							posobj = new FociObject[1]; // if not make array of length 1
							posobj[0] = (distancelist.get(0)).getF(); // add the first object
						} else {
							if ((area[objnum]<three_foci) && (dist.length>1)) { //if there are 2
								posobj = new FociObject[2]; // make an array of 2, and add obj 1 and 2
								posobj[0] = (distancelist.get(0)).getF();
								posobj[1] = (distancelist.get(1)).getF();
							} else { 
								if (dist.length>2 && distsize>2) { // even bigger add 3
									posobj = new FociObject[3];
									posobj[0] = (distancelist.get(0)).getF();
									posobj[1] = (distancelist.get(1)).getF();
									posobj[2] = (distancelist.get(2)).getF();
								}
							}
						}
					} else { // with only one object that is the neighbour
						posobj = new FociObject[1];
						posobj[0] = (distancelist.get(0)).getF();
					 }
					Integer[] indicesOAi = new Integer[0];
					// Track all the objects
					//IJ.log("posobj "+posobj.length); 
					for (int m=0;m<posobj.length;m++){
						//indices = OA.find(iter-spaceindex-1, posobj[m]);
						indicesOAi = OAi.find(iter-spaceindex-1, posobj[m].FOindex);
						if (indicesOAi.length>0){
							//if (indicesOAi.length == 1){ // this happens if this is only one object
					    		//if (!OA.get(iter-1, indices[0])){
						  	if (!OAi.get(iter-1,indicesOAi[0])){
								OAi.set(iter-1, indicesOAi[0], CurrentFoci.FOindex);
							  	OA.set(iter-1,indicesOAi[0],CurrentFoci);
							} else {
								if (spaceindex == 1){
									double x_cur = OA.getreal(iter-1,indicesOAi[0]).x_coord;
									double y_cur = OA.getreal(iter-1,indicesOAi[0]).y_coord;
									double x_1 = OA.getreal(iter-spaceindex-1,indicesOAi[0]).x_coord;
									double y_1 = OA.getreal(iter-spaceindex-1,indicesOAi[0]).y_coord;
									double del_xold = ((x_cur-x_1)*(x_cur-x_1));
									double del_yold = ((y_cur-y_1)*(y_cur-y_1));
									double dist_old = Math.sqrt(del_xold + del_yold);
									if (distancelist.get(m).getdist() < dist_old){
										// Place current values in new position
										OA.add(iter-1,OA.getreal(iter-1,indicesOAi[0]));
										OA.set(iter-1, indicesOAi[0], CurrentFoci);
										OAi.add(iter-1, OA.getreal(iter-1, indicesOAi[0]).FOindex);
										OAi.set(iter-1, indicesOAi[0], CurrentFoci.FOindex);
									} else {
										OA.add(iter-1,CurrentFoci);
										OAi.add(iter-1,CurrentFoci.FOindex);
									}
								} else {
									OA.add(iter-1,CurrentFoci);
									OAi.add(iter-1, CurrentFoci.FOindex);
								}
							}
							// this part is now disabled as it causes crashes on large cells
							/*} else { // Multiple instances to check, meaning the object is so big it likely contains multiple foci
								double theta[];
								theta = new double[10];
								for (int kk=0;kk<indicesOAi.length;kk++){
									double x_now = OA.getreal(iter-spaceindex-1,indicesOAi[kk]).x_coord;
									double y_now = OA.getreal(iter-spaceindex-1,indicesOAi[kk]).y_coord;
									double x_prev = OA.getreal(iter-(spaceindex+1)-1,indicesOAi[kk]).x_coord;
									double y_prev = OA.getreal(iter-(spaceindex+1)-1,indicesOAi[kk]).y_coord;
									double vector1 = Math.sqrt(((x_coord[objnum]-x_now)*(x_coord[objnum]-x_now))+((y_coord[objnum]-y_now)*(y_coord[objnum]-y_now)));
									double vector2 = Math.sqrt(((x_coord[objnum]-x_prev)*(x_coord[objnum]-x_prev))+((y_coord[objnum]-y_prev)*(y_coord[objnum]-y_prev)));
									double dot = ((x_now-x_coord[objnum])*(x_prev-x_coord[objnum]))+((y_now-y_coord[objnum])*(y_prev-y_coord[objnum]));
									theta[kk] = Math.acos(dot/(vector1*vector2));
									theta[kk] = Math.abs(theta[kk]-Math.PI/2);
								}
								double[] thetaori = theta;
								Arrays.sort(theta);
								// Find index
								for (int sorter=0;sorter<theta.length;sorter++){
									if (thetaori[sorter] == theta[0]){
										OA.set(iter-1, indicesOAi[sorter], CurrentFoci);
										OAi.set(iter-1, indicesOAi[sorter], CurrentFoci.FOindex);
									}
								}
							}*/
						} else {
							OA.add(iter-1,CurrentFoci);
							OAi.add(iter-1, CurrentFoci.FOindex);
						}
					}
				}
				frameN=0;
				objnum++;
				if (objnum==frame.length){// stop after going through all objects
					break;
				}
			}
			// Clear space
			//s.empty();
			//s = new space(delx,dely);
			// Make the new space
			FociObject[] toadd = toaddtospace.toArray(new FociObject[toaddtospace.size()]);
			//IJ.log(""+toadd.length);
			for(int i=0; i<toadd.length; i++){
				int ii = toadd[i].ii;
				int jj = toadd[i].jj;
				s.addtospace(ii,jj,toadd[i]);
			}
			OA.resize(iter-1);
			OAi.resize(iter-1);
		}
		IJ.log("Tracking done!");
		PlotDataJoost.plot(OAi,OA, ColorSet, imp);  //edited out but dont know why feb 2016
		PlotData.plot(OAi, OA, Data.directory);
   	//same
	}
}
